# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "python-dotenv==1.0.1",
#     "huggingface-hub==0.24.0",
#     "openai==1.61.1",
#     "pandas==2.2.0",
#     "pydantic==2.10.6",
#     "requests==2.32.3",
# ]
# ///

"""Create the BE-WSC dataset (Winograd Schema Challenge, Belarusian).

This script converts the BelarusianGLUE `bewsc_as_wsc` split into a Winogrande-like
multiple choice dataset.

The source dataset provides one candidate antecedent (`span1_text`) and a label
indicating whether the pronoun (`span2_text`) refers to it. We use an LLM to
extract a second candidate antecedent from the sentence to form a 2-way choice.
"""

import json
import os
import textwrap
from collections import Counter

import pandas as pd
from datasets import Dataset, DatasetDict, Split, disable_progress_bars, load_dataset
from dotenv import load_dotenv
from huggingface_hub import HfApi
from openai import OpenAI
from openai.types.chat import ChatCompletionUserMessageParam
from pydantic import BaseModel

from .constants import (
    CHOICES_MAPPING,
    MAX_NUM_CHARS_IN_INSTRUCTION,
    MAX_NUM_CHARS_IN_OPTION,
    MAX_REPETITIONS,
    MIN_NUM_CHARS_IN_INSTRUCTION,
    MIN_NUM_CHARS_IN_OPTION,
)

load_dotenv()

CACHE_FILE: str = "be_wsc_cache.json"

candidate_cache: dict[str, str] = {}
client: OpenAI | None = None


class SecondCandidate(BaseModel):
    """Structured output: extracted second candidate span from the sentence."""

    distractor: str


def load_cache() -> dict[str, str]:
    """Load cache from CACHE_FILE if it exists.

    Returns:
        The cache.
    """
    try:
        with open(CACHE_FILE, "r") as cache_file:
            return json.load(cache_file)
    except FileNotFoundError:
        return {}


def save_cache(cache: dict) -> None:
    """Save cache to CACHE_FILE.

    Args:
        cache: The cache to save.
    """
    with open(CACHE_FILE, "w") as cache_file:
        json.dump(cache, cache_file, indent=4)


def main() -> None:
    """Create the BE-WSC dataset and upload it to the HF Hub."""
    disable_progress_bars()

    global candidate_cache, client
    candidate_cache = load_cache()
    client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

    repo_id = "maaxap/BelarusianGLUE"

    # Download the dataset
    dataset = load_dataset(path=repo_id, name="bewsc_as_wsc")
    assert isinstance(dataset, DatasetDict)

    # Convert the dataset to a dataframe
    train_df = dataset["train"].to_pandas()
    val_df = dataset["validation"].to_pandas()
    test_df = dataset["test"].to_pandas()
    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    # Check for duplicate idx values (used for cache keys)
    assert train_df["idx"].is_unique, "The train dataframe has duplicate idx values"
    assert val_df["idx"].is_unique, "The validation dataframe has duplicate idx values"
    assert test_df["idx"].is_unique, "The test dataframe has duplicate idx values"

    # Add split names (used for cache keys)
    train_df["split"] = "train"
    val_df["split"] = "validation"
    test_df["split"] = "test"

    train_df = prepare_dataframe(df=train_df)
    val_df = prepare_dataframe(df=val_df)
    test_df = prepare_dataframe(df=test_df)

    train_df, val_df, test_df = make_splits(
        train_df=train_df, val_df=val_df, test_df=test_df
    )

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Push the dataset to the Hugging Face Hub
    target_dataset_id = "EuroEval/be-wsc"

    HfApi().delete_repo(target_dataset_id, repo_type="dataset", missing_ok=True)
    dataset.push_to_hub(target_dataset_id, private=True)


def make_splits(
    train_df: pd.DataFrame, val_df: pd.DataFrame, test_df: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Make splits for the dataset.

    This function creates a train/val split and moves the remaining samples into test.

    Args:
        train_df: The training dataframe.
        val_df: The validation dataframe.
        test_df: The test dataframe.

    Returns:
        The final training, validation, and test dataframes.
    """
    val_size = 64
    train_size = 128

    val_size = min(val_size, len(val_df))
    train_size = min(train_size, len(train_df))

    val_df_sample = val_df.sample(n=val_size, random_state=42)
    val_remaining_df = val_df.drop(val_df_sample.index)

    train_df_sample = train_df.sample(n=train_size, random_state=42)
    train_remaining_df = train_df.drop(train_df_sample.index)

    test_df = pd.concat(
        [train_remaining_df, val_remaining_df, test_df], ignore_index=True
    )

    # Reset indices
    train_df = train_df_sample.reset_index(drop=True)
    val_df = val_df_sample.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)
    return train_df, val_df, test_df


def prepare_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """Prepare the dataframe by filtering and processing.

    Args:
        df:
            Input dataframe.

    Returns:
        Processed dataframe.
    """
    df = df.copy()

    required_cols = {
        "idx",
        "split",
        "text",
        "span1_index",
        "span1_text",
        "span2_index",
        "span2_text",
        "label",
    }
    missing = required_cols.difference(df.columns)
    if missing:
        raise ValueError(f"Missing required columns: {sorted(missing)}")

    # Extract a second candidate referent/antecedent span for each sample
    # (cached, via LLM).
    # - If label==1, this should be a plausible distractor (wrong option).
    # - If label==0, span1 is wrong, so this should be the true referent (right option).
    df["second_candidate"] = df.apply(_extract_second_candidate, axis=1)

    # Drop rows where we failed to extract a usable second candidate
    df = df.loc[df["second_candidate"].astype(str).str.strip().astype(bool)]

    # Build Winogrande-like instruction: replace the pronoun (span2) with a blank "_"
    df["instruction"] = df.apply(_make_instruction, axis=1)

    # Options: A is the provided candidate antecedent (span1), B is the extracted one
    df["option_a"] = df["span1_text"].astype(str)
    df["option_b"] = df["second_candidate"].astype(str)

    # label==1 means span2 refers to span1 -> option_a is correct, else option_b
    df["label"] = df["label"].map(lambda x: "a" if int(x) == 1 else "b")

    # Remove the samples with overly short or long texts/options
    df = df.loc[
        (df.instruction.str.len() >= MIN_NUM_CHARS_IN_INSTRUCTION)
        & (df.instruction.str.len() <= MAX_NUM_CHARS_IN_INSTRUCTION)
        & (df.option_a.str.len() >= MIN_NUM_CHARS_IN_OPTION)
        & (df.option_a.str.len() <= MAX_NUM_CHARS_IN_OPTION)
        & (df.option_b.str.len() >= MIN_NUM_CHARS_IN_OPTION)
        & (df.option_b.str.len() <= MAX_NUM_CHARS_IN_OPTION)
    ]

    # Remove overly repetitive samples
    df = df.loc[
        ~df.instruction.apply(is_repetitive)
        & ~df.option_a.apply(is_repetitive)
        & ~df.option_b.apply(is_repetitive)
    ]

    # Make a `text` column with all the options in it
    df["text"] = [
        row.instruction.replace("\n", " ").strip() + "\n"
        f"{CHOICES_MAPPING['be']}:\n"
        "a. " + row.option_a.replace("\n", " ").strip() + "\n"
        "b. " + row.option_b.replace("\n", " ").strip()
        for _, row in df.iterrows()
    ]

    # Only keep the `text` and `label` columns
    df = df.loc[:, ["text", "label"]]

    # Remove duplicates
    df.drop_duplicates(inplace=True)
    df.reset_index(drop=True, inplace=True)

    return df


def is_repetitive(text: str) -> bool:
    """Return True if the text is repetitive.

    Args:
        text: The text to check for repetition.

    Returns:
        True if the text is repetitive, False otherwise.
    """
    max_repetitions = max(Counter(text.split()).values())
    return max_repetitions > MAX_REPETITIONS


def _cache_key(row: pd.Series) -> str:
    """Create a stable cache key for the second-candidate extraction call.

    Args:
        row: The row of the dataframe.

    Returns:
        The cache key.
    """
    split = str(row["split"]).strip()
    idx = int(row["idx"])
    return f"{split}_{idx}"


def _extract_second_candidate(row: pd.Series) -> str:
    """Extract a second candidate referent span from the sentence (cached).

    Note: this is a "distractor" only when label==1. When label==0, span1 is wrong,
    so the extracted span should be the true referent.

    Args:
        row: The row of the dataframe.

    Returns:
        The second candidate span.
    """
    key = _cache_key(row)
    if key in candidate_cache:
        cached = candidate_cache[key]
        return str(cached)

    text = str(row["text"])
    pronoun = str(row["span2_text"])
    candidate = str(row["span1_text"])
    label = int(row["label"])

    prompt = textwrap.dedent(
        f"""
        You are helping build a Winogrande-style multiple choice dataset in Belarusian.

        Return your response as a JSON object with exactly this schema:
        {{"distractor": string}}.
        Do not include any other keys, text, markdown, or explanation.

        You are given:
        - A Belarusian sentence.
        - A pronoun (or pronominal adverb) that appears in the sentence.
        - A candidate antecedent span from the same sentence.
        - A label: 1 means the pronoun DOES refer to the candidate; 0 means it DOES
          NOT.

        Your task depends on the label:
        - If label = 1: the candidate is the TRUE referent. Extract EXACTLY ONE
          OTHER span from the sentence that could plausibly be a referent (a
          distractor / wrong option).
        - If label = 0: the candidate is a FALSE option. Extract the TRUE referent
          span from the sentence (the thing the pronoun actually refers to).

        Rules (must follow):
        - Return EXACTLY a substring that appears in the sentence (copy-paste).
        - Return a content span (noun phrase / named entity / location phrase), not
          a pronoun.
        - It must NOT be identical to the candidate span.
        - It must NOT be the pronoun itself.
        - Prefer an earlier mention in the sentence when possible.
        - Keep the span short and option-like when possible. In this dataset,
          candidate spans are usually 1 word (~81%) or 2 words (~15%), sometimes 3
          (~3%), and 4+ words < 1%.
          Use a longer span only if needed for a natural referent.
        - If there is no reasonable alternative candidate, return an empty string.

        Few-shot examples (Belarusian):
        Example (label=1, need a distractor / wrong option)
        Sentence: Аліса шукала ў натоўпе сваю сяброўку Надзю. З-за таго, што яна заўсёды
        носіць чырвоны каптур, Аліса хутка заўважыла яе.
        Pronoun: яна
        Candidate: Надзю
        label: 1
        Answer: {{"distractor": "Аліса"}}

        Example (label=1, need a distractor / wrong option)
        Sentence: Дзяніс растлумачыў сваю тэорыю Марку, але ён не пераканаў яго.
        Pronoun: ён
        Candidate: Дзяніс
        label: 1
        Answer: {{"distractor": "Марку"}}

        Example (label=0, need the TRUE referent)
        Sentence: Мужчына ўзяў заплаканага хлопчыка за руку. Яго далонь была
        вялікай і цёплай.
        Pronoun: Яго
        Candidate: заплаканага хлопчыка
        label: 0
        Answer: {{"distractor": "Мужчына"}}

        Example (label=0, need the TRUE referent)
        Sentence: Доктарка паведаміла Кацярыне, што яна сыходзіць на пенсію, і
        прапанавала
        некалькі варыянтаў працягу лячэння.
        Pronoun: яна
        Candidate: Кацярыне
        label: 0
        Answer: {{"distractor": "Доктарка"}}

        Example (label=1, pronominal adverb)
        Sentence: Мы вырашылі правесці дзень на возеры, бо на ўзбярэжжы акіяна
        бачылі акулу,
        так што плаваць там было небяспечна.
        Pronoun: там
        Candidate: на ўзбярэжжы акіяна
        label: 1
        Answer: {{"distractor": "на возеры"}}

        Now do the same for this input. Remember: output ONLY valid JSON like
        {{"distractor": "..."}}

        Sentence: <text>{text}</text>
        Pronoun: <pronoun>{pronoun}</pronoun>
        Candidate: <candidate>{candidate}</candidate>
        label: <label>{label}</label>
        """
    ).strip()

    messages: list[ChatCompletionUserMessageParam] = [
        ChatCompletionUserMessageParam(role="user", content=prompt)
    ]

    completion = client.beta.chat.completions.parse(
        model="gpt-4o", messages=messages, response_format=SecondCandidate
    )
    parsed = completion.choices[0].message.parsed
    if parsed is None:
        raise ValueError("Parsed response is None")

    second_candidate = parsed.distractor.strip()
    if second_candidate and second_candidate not in text:
        # Enforce "exact substring" to avoid downstream mismatches.
        second_candidate = ""

    if second_candidate and second_candidate == candidate:
        second_candidate = ""
    if second_candidate and second_candidate == pronoun:
        second_candidate = ""

    candidate_cache[key] = second_candidate
    save_cache(cache=candidate_cache)
    return second_candidate


def _mask_pronoun(text: str, pronoun: str, word_index: int) -> str:
    """Replace the pronoun token (by word index) with '_'.

    Args:
        text: The text to mask the pronoun in.
        pronoun: The pronoun to mask.
        word_index: The index of the pronoun in the text.

    Returns:
        The text with the pronoun masked.
    """
    words = text.split()
    assert 0 <= word_index < len(words), "Word index is out of range"

    token = words[word_index]
    if token == pronoun:
        words[word_index] = "_"
    elif pronoun and pronoun in token:
        # Keep punctuation attached to the token (e.g., "яна," -> "_,")
        words[word_index] = token.replace(pronoun, "_", 1)
    else:
        assert False, f"Pronoun not found in token: {token} in text: {text}"

    return " ".join(words)


def _make_instruction(row: pd.Series) -> str:
    """Create the instruction string in Winogrande-like format.

    Args:
        row: The row of the dataframe.

    Returns:
        The instruction string.
    """
    text = str(row["text"]).replace("\n", " ").strip()
    pronoun = str(row["span2_text"]).strip()
    pronoun_idx = int(row["span2_index"])
    masked = _mask_pronoun(text=text, pronoun=pronoun, word_index=pronoun_idx)
    return f"{masked} Да каго або чаго адносіцца пропуск _?"


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the PoQuaD-mini dataset and upload it to the HF Hub."""

import pandas as pd
from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict
from datasets.load import load_dataset
from datasets.splits import Split
from huggingface_hub.hf_api import HfApi
from requests.exceptions import HTTPError

from .constants import (
    MAX_NUM_CHARS_IN_CONTEXT,
    MAX_NUM_CHARS_IN_QUESTION,
    MIN_NUM_CHARS_IN_CONTEXT,
    MIN_NUM_CHARS_IN_QUESTION,
)


def main() -> None:
    """Create the PoQuaD-mini dataset and upload it to the HF Hub."""
    dataset_id = "mzasada/poquad_v2"

    # Load the dataset
    train = load_dataset(dataset_id, split="train")
    val = load_dataset(dataset_id, split="validation")

    # Ensure that the datasets are indeed datasets
    assert isinstance(train, Dataset)
    assert isinstance(val, Dataset)

    train_df = train.to_pandas()
    val_df = val.to_pandas()
    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)

    # Merge the splits
    df = pd.concat([train_df, val_df], ignore_index=True)

    # Ignore impossible questions
    df = df[df.is_impossible == "False"]

    # Ensure that `df` is indeed a Pandas DataFrame
    assert isinstance(df, pd.DataFrame)

    # Only work with samples where the context is not very large or small
    lengths = df.context.str.len()
    df = df[lengths.between(MIN_NUM_CHARS_IN_CONTEXT, MAX_NUM_CHARS_IN_CONTEXT)]

    # Only work with samples where the question is not very large or small
    lengths = df.question.str.len()
    df_with_no_outliers = df[
        lengths.between(MIN_NUM_CHARS_IN_QUESTION, MAX_NUM_CHARS_IN_QUESTION)
    ]
    assert isinstance(df_with_no_outliers, pd.DataFrame)

    # Only work with the questions having answers in the context
    has_answer: pd.Series = df_with_no_outliers.answers.map(
        lambda dct: dct["text"][0] != ""
    )
    df_with_answer: pd.DataFrame = df_with_no_outliers.loc[has_answer]

    # Create validation split
    val_size = 256
    val_df = df_with_answer.sample(n=val_size, random_state=4242)

    # Create test split
    test_size = 2048
    df_with_answer_filtered: pd.DataFrame = df_with_answer.loc[
        ~df_with_answer.index.isin(val_df.index)
    ]
    test_df = df_with_answer_filtered.sample(n=test_size, random_state=4242)

    # Create train split
    train_size = 1024
    full_train_df_with_answer = df_with_answer_filtered.loc[
        ~df_with_answer_filtered.index.isin(test_df.index)
    ]
    train_df = full_train_df_with_answer.sample(n=train_size, random_state=4242)

    val_df = val_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)
    train_df = train_df.reset_index(drop=True)

    # Collect datasets in a dataset dictionary
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    mini_dataset_id = "EuroEval/poquad-mini"

    # Remove the dataset from Hugging Face Hub if it already exists
    try:
        api: HfApi = HfApi()
        api.delete_repo(mini_dataset_id, repo_type="dataset")
    except HTTPError:
        pass

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(mini_dataset_id, private=True)


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==4.0.0",
#     "huggingface-hub==0.34.4",
#     "requests==2.32.5",
# ]
# ///

"""Create the Estonian Winogrande dataset and upload to HF Hub."""

from typing import MutableMapping

from datasets import Dataset, DatasetDict, load_dataset
from huggingface_hub import HfApi

from .constants import CHOICES_MAPPING


def main() -> None:
    """Create the Estonian Winogrande dataset and upload to HF Hub."""
    target_repo_id = "EuroEval/winogrande-et"

    # start from the official source
    human_ds = load_dataset("tartuNLP/winogrande_et", "human_translated")
    mt_ds = load_dataset("tartuNLP/winogrande_et", "machine_translated")
    assert isinstance(human_ds, DatasetDict)
    assert isinstance(mt_ds, DatasetDict)

    # target split sizes
    train_size = 1024
    val_size = 256
    test_size = 2048

    train = mt_ds["train"].select(range(train_size))
    val = mt_ds["dev"].select(range(val_size))
    test = human_ds["test"].select(range(min(test_size, len(human_ds["test"]))))

    assert isinstance(train, Dataset)
    assert isinstance(val, Dataset)
    assert isinstance(test, Dataset)

    # we don't have human translations for train and dev
    ds = DatasetDict({"train": train, "val": val, "test": test})

    # please don't share the answers explicitly though
    mapped_test = ds["test"].map(lambda row: {"answer": row["qID"][-1]})
    assert isinstance(mapped_test, type(ds["test"]))
    ds["test"] = mapped_test

    # add options to the text and transform labels to letters
    ds = ds.map(add_options_and_label)

    # retain only used columns
    ds = ds.select_columns(["text", "label"])

    # Push the dataset to the Hugging Face Hub
    HfApi().delete_repo(target_repo_id, repo_type="dataset", missing_ok=True)
    ds.push_to_hub(target_repo_id, private=True)


def add_options_and_label(row: MutableMapping) -> MutableMapping:
    """Add options to the text and transform labels to letters.

    Args:
        row:
            A row from the dataset.

    Returns:
        A dictionary with the modified text and label.
    """
    letter_mapping = {"1": "a", "2": "b"}

    original_text = row["sentence"]
    option_1 = row["option1"]
    option_2 = row["option2"]

    new_text = (
        f"{original_text}\n{CHOICES_MAPPING['et']}:\na. {option_1}\nb. {option_2}"
    )

    answer = row["answer"]
    if answer not in letter_mapping.keys():
        raise ValueError(
            f"Invalid answer: {answer}, possible values are {letter_mapping.keys()}"
        )
    label = letter_mapping[answer]

    return {"text": new_text, "label": label}


if __name__ == "__main__":
    main()



"""Load custom dataset configs."""

import importlib.util
import logging
from pathlib import Path
from types import ModuleType

from huggingface_hub import HfApi

from .data_models import DatasetConfig
from .logging_utils import log_once
from .utils import get_hf_token


def load_custom_datasets_module(custom_datasets_file: Path) -> ModuleType | None:
    """Load the custom datasets module if it exists.

    Args:
        custom_datasets_file:
            The path to the custom datasets module.

    Raises:
        RuntimeError:
            If the custom datasets module cannot be loaded.
    """
    if custom_datasets_file.exists():
        spec = importlib.util.spec_from_file_location(
            name="custom_datasets_module", location=str(custom_datasets_file.resolve())
        )
        if spec is None:
            log_once(
                "Could not load the spec for the custom datasets file from "
                f"{custom_datasets_file.resolve()}.",
                level=logging.ERROR,
            )
            return None
        module = importlib.util.module_from_spec(spec=spec)
        if spec.loader is None:
            log_once(
                "Could not load the module for the custom datasets file from "
                f"{custom_datasets_file.resolve()}.",
                level=logging.ERROR,
            )
            return None
        spec.loader.exec_module(module)
        return module
    return None


def try_get_dataset_config_from_repo(
    dataset_id: str, api_key: str | None, cache_dir: Path
) -> DatasetConfig | None:
    """Try to get a dataset config from a Hugging Face dataset repository.

    Args:
        dataset_id:
            The ID of the dataset to get the config for.
        api_key:
            The Hugging Face API key to use to check if the repositories have custom
            dataset configs.
        cache_dir:
            The directory to store the cache in.

    Returns:
        The dataset config if it exists, otherwise None.
    """
    # Check if the dataset ID is a Hugging Face dataset ID, abort if it isn't
    token = get_hf_token(api_key=api_key)
    hf_api = HfApi(token=token)
    if not hf_api.repo_exists(repo_id=dataset_id, repo_type="dataset"):
        return None

    # Check if the repository has a euroeval_config.py file, abort if it doesn't
    repo_files = hf_api.list_repo_files(
        repo_id=dataset_id, repo_type="dataset", revision="main"
    )
    if "euroeval_config.py" not in repo_files:
        log_once(
            f"Dataset {dataset_id} does not have a euroeval_config.py file, so we "
            "cannot load it. Skipping.",
            level=logging.WARNING,
        )
        return None

    # Fetch the euroeval_config.py file, abort if loading failed
    external_config_path = cache_dir / "external_dataset_configs" / dataset_id
    external_config_path.mkdir(parents=True, exist_ok=True)
    hf_api.hf_hub_download(
        repo_id=dataset_id,
        repo_type="dataset",
        filename="euroeval_config.py",
        local_dir=external_config_path,
        local_dir_use_symlinks=False,
    )
    module = load_custom_datasets_module(
        custom_datasets_file=external_config_path / "euroeval_config.py"
    )
    if module is None:
        return None

    # Check that there is exactly one dataset config, abort if there isn't
    repo_dataset_configs = [
        cfg for cfg in vars(module).values() if isinstance(cfg, DatasetConfig)
    ]
    if not repo_dataset_configs:
        return None  # Already warned the user in this case, so we just skip
    elif len(repo_dataset_configs) > 1:
        log_once(
            f"Dataset {dataset_id} has multiple dataset configurations. Please ensure "
            "that only a single DatasetConfig is defined in the `euroeval_config.py` "
            "file.",
            level=logging.WARNING,
        )
        return None

    # Get the dataset split names
    splits = [
        split["name"]
        for split in hf_api.dataset_info(repo_id=dataset_id).card_data.dataset_info[
            "splits"
        ]
    ]
    train_split_candidates = sorted(
        [split for split in splits if "train" in split.lower()], key=len
    )
    val_split_candidates = sorted(
        [split for split in splits if "val" in split.lower()], key=len
    )
    test_split_candidates = sorted(
        [split for split in splits if "test" in split.lower()], key=len
    )
    train_split = train_split_candidates[0] if train_split_candidates else None
    val_split = val_split_candidates[0] if val_split_candidates else None
    test_split = test_split_candidates[0] if test_split_candidates else None
    if test_split is None:
        log_once(
            f"Dataset {dataset_id} does not have a test split, so we cannot load it. "
            "Please ensure that the dataset has a test split.",
            level=logging.ERROR,
        )
        return None

    # Set up the config with the repo information
    repo_dataset_config = repo_dataset_configs[0]
    repo_dataset_config.name = dataset_id
    repo_dataset_config.pretty_name = dataset_id
    repo_dataset_config.source = dataset_id
    repo_dataset_config.train_split = train_split
    repo_dataset_config.val_split = val_split
    repo_dataset_config.test_split = test_split

    return repo_dataset_config



"""All Latvian dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import LATVIAN
from ..tasks import COMMON_SENSE, KNOW, LA, NER, RC, SENT, SUMM

### Official datasets ###

LATVIAN_TWITTER_SENTIMENT_CONFIG = DatasetConfig(
    name="latvian-twitter-sentiment",
    pretty_name="Latvian Twitter Sentiment",
    source="EuroEval/latvian-twitter-sentiment-mini",
    task=SENT,
    languages=[LATVIAN],
)

SCALA_LV_CONFIG = DatasetConfig(
    name="scala-lv",
    pretty_name="ScaLA-lv",
    source="EuroEval/scala-lv",
    task=LA,
    languages=[LATVIAN],
)

FULLSTACK_NER_LV_CONFIG = DatasetConfig(
    name="fullstack-ner-lv",
    pretty_name="FullStack NER-lv",
    source="EuroEval/fullstack-ner-lv-mini",
    task=NER,
    languages=[LATVIAN],
)

MULTI_WIKI_QA_LV_CONFIG = DatasetConfig(
    name="multi-wiki-qa-lv",
    pretty_name="MultiWikiQA-lv",
    source="EuroEval/multi-wiki-qa-lv-mini",
    task=RC,
    languages=[LATVIAN],
)

LSM_CONFIG = DatasetConfig(
    name="lsm",
    pretty_name="LSM",
    source="EuroEval/lsm-mini",
    task=SUMM,
    languages=[LATVIAN],
)


MMLU_LV_CONFIG = DatasetConfig(
    name="mmlu-lv",
    pretty_name="MMLU-lv",
    source="EuroEval/mmlu-lv-mini",
    task=KNOW,
    languages=[LATVIAN],
)

COPA_LV_CONFIG = DatasetConfig(
    name="copa-lv",
    pretty_name="COPA-lv",
    source="EuroEval/copa-lv",
    task=COMMON_SENSE,
    languages=[LATVIAN],
    labels=["a", "b"],
)


### Unofficial datasets ###

WIKIANN_LV_CONFIG = DatasetConfig(
    name="wikiann-lv",
    pretty_name="WikiANN-lv",
    source="EuroEval/wikiann-lv-mini",
    task=NER,
    languages=[LATVIAN],
    unofficial=True,
)

WINOGRANDE_LV_CONFIG = DatasetConfig(
    name="winogrande-lv",
    pretty_name="Winogrande-lv",
    source="EuroEval/winogrande-lv",
    task=COMMON_SENSE,
    languages=[LATVIAN],
    labels=["a", "b"],
    unofficial=True,
)



"""Utility functions related to the sequence-classification task group."""

import collections.abc as c
import logging
import re
import typing as t

import Levenshtein
import numpy as np

from ..enums import TaskGroup
from ..exceptions import InvalidBenchmark
from ..string_utils import extract_multiple_choice_labels
from ..types import Predictions
from ..utils import log_once, raise_if_model_output_contains_nan_values

if t.TYPE_CHECKING:
    from datasets.arrow_dataset import Dataset
    from transformers.trainer_utils import EvalPrediction

    from ..data_models import (
        BenchmarkConfig,
        DatasetConfig,
        GenerativeModelOutput,
        ModelConfig,
    )
    from ..types import Labels


def compute_metrics(
    model_outputs_and_labels: "tuple[Predictions, Labels] | EvalPrediction",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
    dataset: "Dataset",
) -> dict[str, float]:
    """Compute the metrics needed for evaluation.

    Args:
        model_outputs_and_labels:
            The first sequence contains the model outputs and the second sequence
            contains the true labels.
        dataset_config:
            The configuration of the dataset.
        benchmark_config:
            The configuration of the benchmark.
        dataset:
            The dataset used for evaluation. This is only used in case any additional
            metadata is used to compute the metrics.

    Returns:
        A dictionary with the names of the metrics as keys and the metric values as
        values.
    """
    model_outputs, labels = model_outputs_and_labels
    label2id = {label: idx for idx, label in dataset_config.id2label.items()}

    # If the model outputs is a pair, then the first element corresponds to the model
    # predictions
    if isinstance(model_outputs, tuple) and len(model_outputs) == 2:
        model_outputs = model_outputs[0]

    model_output_dtype = np.asarray(model_outputs).dtype
    if model_output_dtype in [np.float16, np.float32, np.float64]:
        predictions = np.asarray(model_outputs).argmax(axis=-1)
    else:
        predictions = model_outputs

    raise_if_model_output_contains_nan_values(model_output=model_outputs)  # type: ignore[bad-argument-type]

    prompt_label_to_label_mapping = {
        prompt_label: label
        for label, prompt_label in dataset_config.prompt_label_mapping.items()
    }
    predictions = [
        (
            label2id[prompt_label_to_label_mapping[pred.lower()]]
            if isinstance(pred, str)
            else pred
        )
        for pred in predictions  # type: ignore[not-iterable]
    ]

    label_ids = [
        label2id[label.lower()] if isinstance(label, str) else label for label in labels
    ]

    results: dict[str, float] = dict()
    for metric in dataset_config.task.metrics:
        score: float | None = metric(
            predictions=predictions,
            references=label_ids,
            dataset=dataset,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
        )

        # The metric returns None if we are running on multi-GPU and the current
        # process is not the main process
        if score is not None:
            results[metric.name] = score

    return results


def extract_labels_from_generation(
    input_batch: dict[str, list],
    model_output: "GenerativeModelOutput",
    dataset_config: "DatasetConfig",
    model_config: "ModelConfig",
    first_label_token_mapping: dict[str, str] | bool,
) -> c.Sequence[str]:
    """Extract the predicted labels from the generated output.

    Args:
        input_batch:
            The input batch, where the keys are the feature names and the values
            are lists with the feature values.
        model_output:
            The raw generated output of the model.
        dataset_config:
            The configuration of the dataset.
        model_config:
            The configuration of the model.
        first_label_token_mapping:
            A mapping from labels to the first token in each label, or alternatively a
            Boolean value indicating whether the model should output scores (if the
            mapping is outputted then the model will always output scores).

    Returns:
        The predicted labels.

    Raises:
        InvalidBenchmark:
            If the task requires log probabilities, but the model did not output them,
            or if the model outputted log probabilities but the first label token
            mapping is not provided.
    """
    # Get the candidate labels, which are the labels that the model can predict
    default_labels = [
        dataset_config.prompt_label_mapping[lbl]
        for lbl in dataset_config.id2label.values()
    ]
    if dataset_config.task.task_group == TaskGroup.MULTIPLE_CHOICE_CLASSIFICATION:
        sample_candidate_labels = [
            extract_multiple_choice_labels(
                prompt=prompt, candidate_labels=default_labels
            )
            for prompt in input_batch["prompt"]
        ]
    else:
        sample_candidate_labels = [default_labels] * len(input_batch["prompt"])

    if model_output.scores is not None:
        if first_label_token_mapping is False:
            raise InvalidBenchmark(
                "The model outputted logprobs, but the first label token mapping is "
                "not provided, which is not supported."
            )
        labels = get_closest_logprobs_labels(
            generation_logprobs=model_output.scores,
            first_label_token_mapping=first_label_token_mapping,
            candidate_labels=sample_candidate_labels,
        )
        if labels is not None:
            return labels
        elif dataset_config.task.requires_logprobs:
            raise InvalidBenchmark(
                "This task requires the model to output logprobs, and this model "
                "does not seem to be able to do that. Skipping the evaluation."
            )

    new_predicted_labels: list[str] = list()
    num_predictions_being_very_off = 0
    for idx, predicted_label in enumerate(model_output.sequences):
        # If the prediction includes a boxed answer, use that instead of the full
        # generation
        if (m := re.search(r"boxed\{(.*?)\}", predicted_label)) is not None:
            predicted_label = m.group(1)

        # If the prediction starts with one of the candidate labels (case-insensitive)
        # then use that one
        prefix_candidate_labels = [
            candidate_label
            for candidate_label in sample_candidate_labels[idx]
            if predicted_label.lower().startswith(candidate_label.lower())
        ]
        if prefix_candidate_labels:
            new_predicted_labels.append(prefix_candidate_labels[0])
            continue

        # We set the word edit distance weights such that we heavily penalise insertions
        # and substitutions, so that we don't just insert the correct label, but that we
        # want the model to have included the correct label in its output.
        insertion_weight = 1000
        deletion_weight = 1
        substitution_weight = 1000

        # Compute the word edit distances between the predicted label and all candidate
        # labels
        edit_distances = [
            Levenshtein.distance(
                s1=predicted_label.lower(),
                s2=candidate_label.lower(),
                weights=(insertion_weight, deletion_weight, substitution_weight),
            )
            for candidate_label in sample_candidate_labels[idx]
        ]

        best_candidate_label = sample_candidate_labels[idx][
            np.argmin(edit_distances).item()
        ]

        # If no candidate labels were found, we either pick the label with the smallest
        # word edit distance to the predicted label (if invalid model outputs are
        # allowed), or we raise an error
        if min(edit_distances) >= 1000:
            num_predictions_being_very_off += 1

        new_predicted_labels.append(best_candidate_label)

    if num_predictions_being_very_off > 0:
        if dataset_config.allow_invalid_model_outputs:
            log_msg = (
                "No candidate labels found for the predicted label in "
                f"{num_predictions_being_very_off:,}/{len(model_output.sequences):,} "
                f"of the samples with the model {model_config.model_id!r}. This "
                "likely means that the model were completely off in these cases, "
                "but since invalid model outputs are allowed for this task, we used "
                "the closest candidate labels as the output labels."
            )
            level = logging.DEBUG
            if num_predictions_being_very_off / len(model_output.sequences) > 0.5:
                log_msg += (
                    " Since this happened for most of the model's predictions, please "
                    "report this issue to the EuroEval team at "
                    "github.com/EuroEval/EuroEval/issues."
                )
                level = logging.WARNING
            log_once(log_msg, level=level)
        else:
            raise InvalidBenchmark(
                "No candidate labels found for the predicted label in "
                f"{num_predictions_being_very_off:,}/{len(model_output.sequences):,} "
                "of the samples. This likely means that the model were completely "
                "off in these cases. Since this task does not allow invalid model "
                "outputs, we have to abort the evaluation."
            )

    return new_predicted_labels


def get_closest_logprobs_labels(
    generation_logprobs: c.Sequence[c.Sequence[c.Sequence[tuple[str, float]]]],
    first_label_token_mapping: dict[str, str] | t.Literal[True],
    candidate_labels: c.Sequence[c.Sequence[str]],
) -> c.Sequence[str] | None:
    """Get the labels with the highest predicted logprob value.

    In case a candidate label is split into multiple tokens, we only use the first
    token to compute the logprob value. E.g., if the candidate label "positive" is
    tokenised as ["pos", "itive"], we only use the logprob value of "pos" to
    represent the logprob value of the entire label.

    Args:
        generation_logprobs:
            The logprobs of the generated tokens, for all samples in the batch. Of shape
            (batch_size, num_tokens, num_logprobs).
        first_label_token_mapping:
            A mapping from labels to the first token in each label, or alternatively a
            `True` value indicating that the model should output logprobs.
        candidate_labels:
            The candidate labels for each sample in the batch.

    Returns:
        The predicted labels, or None if labels could not be extracted.

    Raises:
        InvalidBenchmark:
            If no candidate label can be found for any of the generated labels.
    """
    output_labels: list[str] = list()
    for idx, sample in enumerate(generation_logprobs):
        for logprob_list in sample:
            generated_labels = [
                re.sub(pattern=r"^[^a-zæøåüöä0-9]+$", repl="", string=label.lower())
                for label, _ in logprob_list
            ]
            generated_labels = [label for label in generated_labels if label != ""]

            # We want to use the first generated label which contains a unique candidate
            # label, as the output label
            output_label: str | None = None
            for generated_label in generated_labels:
                # Get the candidate labels. If we have a first label token mapping, we
                # use it to get the candidate labels. Otherwise, we check if any of the
                # labels start with the generated label.
                if isinstance(first_label_token_mapping, dict):
                    if any(
                        candidate_label not in first_label_token_mapping
                        for candidate_label in candidate_labels[idx]
                    ):
                        raise InvalidBenchmark(
                            "There is a label not present in the first label token "
                            "mapping - this should never happen! Please report this "
                            "issue to the EuroEval team at "
                            "github.com/EuroEval/EuroEval/issues."
                        )

                    candidate_output_labels = {
                        candidate_label
                        for candidate_label in candidate_labels[idx]
                        if generated_label == first_label_token_mapping[candidate_label]
                    }
                else:
                    candidate_output_labels = {
                        candidate_label
                        for candidate_label in candidate_labels[idx]
                        if candidate_label.startswith(generated_label.strip())
                    }

                # If we can uniquely determine the output label, we break the loop.
                if len(candidate_output_labels) == 1:
                    output_label = candidate_output_labels.pop()
                    break

                # If we have multiple candidate labels, we cannot uniquely determine the
                # output label, so we abandon extracting the labels using logprobs and
                # fall back to using word edit distance.
                elif len(candidate_output_labels) > 1:
                    log_once(
                        "Multiple candidate labels found for the generated label "
                        f"{generated_label!r}: {candidate_output_labels}. This means "
                        "that using logprobs to extract the labels is not reliable, "
                        "and we will instead fall back to extracting the labels "
                        "using word edit distance.",
                        level=logging.DEBUG,
                    )
                    return None

                # If no candidate label is found, we first check if any of the labels
                # start with the generated label. This could be the case if the labels
                # in the first token mapping is inaccurate or incomplete, for instance
                # if 'pos' is in the first label token mapping, but the model outputted
                # 'posit'. If this is the case then we cannot trust the first label
                # token mapping, and we fall back to using word edit distance.
                # Otherwise, the generated label is just bad, and we skip to the next
                # generated label.
                elif len(candidate_output_labels) == 0:
                    candidate_output_labels_starting_with_generated_label = [
                        candidate_label
                        for candidate_label in candidate_labels[idx]
                        if candidate_label.startswith(generated_label)
                    ]
                    if candidate_output_labels_starting_with_generated_label:
                        log_once(
                            f"No candidate label found for the generated label "
                            f"{generated_label!r}, but there are candidate labels "
                            f"starting with it: "
                            f"{candidate_output_labels_starting_with_generated_label}. "
                            "This means that the first label token mapping is not "
                            "reliable, and we will instead fall back to extracting "
                            "the labels using word edit distance.",
                            level=logging.DEBUG,
                        )
                        return None

            if output_label is not None:
                output_labels.append(output_label)
                break
        else:
            if len(sample) == 0:
                log_once(
                    "The model outputted an empty string, so no candidate labels could "
                    "be determined. This means that using logprobs to extract the "
                    "labels is not reliable, and we will instead fall back to "
                    "extracting the labels using word edit distance.",
                    level=logging.DEBUG,
                )
            else:
                log_once(
                    "No candidate label found for any of the generated labels, which "
                    "means that using logprobs to extract the labels is not reliable, "
                    "and we will instead fall back to extracting the labels using "
                    "word edit distance.",
                    level=logging.DEBUG,
                )
            return None

    assert len(output_labels) == len(generation_logprobs)
    return output_labels



"""Utility functions related to string manipulation or structuring."""

import collections.abc as c
import logging
import re
import typing as t

import demjson3
import numpy as np

from .exceptions import InvalidBenchmark, InvalidModel
from .logging_utils import log

if t.TYPE_CHECKING:
    from .data_models import ModelIdComponents


def scramble(text: str) -> str:
    """Scramble a string in a bijective manner.

    Args:
        text:
            The string to scramble.

    Returns:
        The scrambled string.
    """
    rng = np.random.default_rng(seed=4242)
    permutation = rng.permutation(x=len(text))
    scrambled = "".join(text[i] for i in permutation)
    return scrambled


def unscramble(scrambled_text: str) -> str:
    """Unscramble a string in a bijective manner.

    Args:
        scrambled_text:
            The scrambled string to unscramble.

    Returns:
        The unscrambled string.
    """
    rng = np.random.default_rng(seed=4242)
    permutation = rng.permutation(x=len(scrambled_text))
    inverse_permutation = np.argsort(permutation)
    unscrambled = "".join(scrambled_text[i] for i in inverse_permutation)
    return unscrambled


def extract_json_dict_from_string(s: str) -> dict | None:
    """Extract a JSON dictionary from a string.

    Args:
        s:
            The string to extract the JSON dictionary from.

    Returns:
        The extracted JSON dictionary, or None if no JSON dictionary could be found.
    """
    json_regex = r"\{[^{}]*?\}"
    if (json_match := re.search(pattern=json_regex, string=s, flags=re.DOTALL)) is None:
        log(
            "The model output does not contain any JSON dictionary, so cannot parse "
            f"it. Skipping. Here is the output: {s!r}",
            level=logging.DEBUG,
        )
        return None
    json_string = json_match.group()
    try:
        json_output = demjson3.decode(txt=json_string)
    except demjson3.JSONDecodeError:
        log(
            "The model output is not valid JSON, so cannot parse it. Skipping. "
            f"Here is the output: {json_string!r}",
            level=logging.DEBUG,
        )
        return None
    if not isinstance(json_output, dict):
        log(
            "The model output is not a JSON dictionary, so cannot parse "
            f"it. Skipping. Here is the output: {json_string!r}",
            level=logging.DEBUG,
        )
        return None
    elif not all(isinstance(key, str) for key in json_output.keys()):
        log(
            "The model output is not a JSON dictionary with string keys, "
            "so cannot parse it. Skipping. Here is the output: "
            f"{json_string!r}",
            level=logging.DEBUG,
        )
        return None
    return json_output


def extract_multiple_choice_labels(
    prompt: str, candidate_labels: c.Sequence[str]
) -> c.Sequence[str]:
    """Extract multiple choice labels from a prompt.

    Args:
        prompt:
            The prompt to extract the labels from.
        candidate_labels:
            The candidate labels to look for in the prompt.

    Returns:
        The extracted labels.
    """
    sample_candidate_labels: list[str] = list()
    for candidate_label in candidate_labels:
        candidate_label_match = re.search(
            pattern=rf"\b{candidate_label}\. ", string=prompt, flags=re.IGNORECASE
        )
        if candidate_label_match is not None:
            sample_candidate_labels.append(candidate_label)
    if not sample_candidate_labels:
        raise InvalidBenchmark(
            "Could not extract any candidate labels from the prompt. Please ensure "
            "that the candidate labels are present in the prompt, each followed by a "
            "dot and a space (e.g., 'a. '). The candidate labels are: "
            f"{', '.join(candidate_labels)}. Here is the prompt: {prompt!r}"
        )
    return sample_candidate_labels


def split_model_id(model_id: str) -> "ModelIdComponents":
    """Split a model ID into its components.

    Args:
        model_id:
            The model ID to split.

    Returns:
        The split model ID.

    Raises:
        If the model ID is not valid.
    """
    # Importing here to avoid circular imports
    from .data_models import ModelIdComponents

    # Attempt to extract the model ID, revision, and param using regex
    model_id_match = re.match(pattern=r"^[^@#]+", string=model_id)
    revision_match = re.search(pattern=r"@([^@#]+)", string=model_id)
    param_match = re.search(pattern=r"#([^@#]+)", string=model_id)

    # If we cannot extract the model ID, raise an error
    if model_id_match is None:
        raise InvalidModel(f"The model ID {model_id!r} is not valid.")
    model_id = model_id_match.group()

    # Extract the revision and param and return the result
    revision = revision_match.group(1) if revision_match is not None else "main"
    param = param_match.group(1) if param_match is not None else None
    return ModelIdComponents(model_id=model_id, revision=revision, param=param)



"""Functions related to the loading of the data."""

import collections.abc as c
import logging
import sys
import time
import typing as t

import requests
from datasets import DatasetDict, load_dataset
from datasets.exceptions import DatasetsError
from huggingface_hub.errors import HfHubHTTPError, RepositoryNotFoundError
from numpy.random import Generator

from .constants import SUPPORTED_FILE_FORMATS_FOR_LOCAL_DATASETS
from .exceptions import HuggingFaceHubDown, InvalidBenchmark
from .logging_utils import log, no_terminal_output
from .string_utils import unscramble
from .tasks import EUROPEAN_VALUES
from .utils import get_hf_token

if t.TYPE_CHECKING:
    from datasets import Dataset

    from .data_models import BenchmarkConfig, DatasetConfig


def load_data(
    rng: Generator, dataset_config: "DatasetConfig", benchmark_config: "BenchmarkConfig"
) -> list["DatasetDict"]:
    """Load the raw bootstrapped datasets.

    Args:
        rng:
            The random number generator to use.
        dataset_config:
            The configuration for the dataset.
        benchmark_config:
            The configuration for the benchmark.

    Returns:
        A list of bootstrapped datasets, one for each iteration.

    Raises:
        InvalidBenchmark:
            If the dataset cannot be loaded.
        HuggingFaceHubDown:
            If the Hugging Face Hub is down.
    """
    dataset = load_raw_data(
        dataset_config=dataset_config,
        cache_dir=benchmark_config.cache_dir,
        api_key=benchmark_config.api_key,
    )

    if (
        not benchmark_config.evaluate_test_split
        and dataset_config.val_split is not None
    ):
        dataset[dataset_config.test_split] = dataset[dataset_config.val_split]

    splits = [
        split
        for split in [
            dataset_config.train_split,
            dataset_config.val_split,
            dataset_config.test_split,
        ]
        if split is not None
    ]

    # Remove empty examples from the datasets
    for text_feature in ["tokens", "text"]:
        for split in splits:
            if text_feature in dataset[split].features:
                dataset = dataset.filter(lambda x: len(x[text_feature]) > 0)

    # If we are testing then truncate the test set, unless we need the full set for
    # evaluation
    if hasattr(sys, "_called_from_test") and dataset_config.task != EUROPEAN_VALUES:
        dataset["test"] = dataset["test"].select(range(1))  # type: ignore[unsupported-operation]

    # Bootstrap the splits, if applicable
    if dataset_config.bootstrap_samples:
        bootstrapped_splits: dict[str, c.Sequence["Dataset"]] = dict()
        for split in splits:
            bootstrap_indices = rng.integers(
                0,
                len(dataset[split]),
                size=(benchmark_config.num_iterations, len(dataset[split])),
            )
            bootstrapped_splits[split] = [  # type: ignore[unsupported-operation]
                dataset[split].select(bootstrap_indices[idx])
                for idx in range(benchmark_config.num_iterations)
            ]
        datasets = [
            DatasetDict(  # type: ignore[no-matching-overload]
                {
                    split: bootstrapped_splits[split][idx]
                    for split in [
                        dataset_config.train_split,
                        dataset_config.val_split,
                        dataset_config.test_split,
                    ]
                    if split is not None
                }
            )
            for idx in range(benchmark_config.num_iterations)
        ]
    else:
        datasets = [dataset] * benchmark_config.num_iterations

    return datasets


def load_raw_data(
    dataset_config: "DatasetConfig", cache_dir: str, api_key: str | None
) -> "DatasetDict":
    """Load the raw dataset.

    Args:
        dataset_config:
            The configuration for the dataset.
        cache_dir:
            The directory to cache the dataset.
        api_key:
            The API key to use as the Hugging Face token.

    Returns:
        The dataset.
    """
    # Case where the dataset source is a Hugging Face ID
    if isinstance(dataset_config.source, str):
        num_attempts = 5
        for _ in range(num_attempts):
            try:
                with no_terminal_output():
                    dataset = load_dataset(
                        path=dataset_config.source.split("::")[0],
                        name=(
                            dataset_config.source.split("::")[1]
                            if "::" in dataset_config.source
                            else None
                        ),
                        cache_dir=cache_dir,
                        token=unscramble("XbjeOLhwebEaSaDUMqqaPaPIhgOcyOfDpGnX_"),
                    )
                break
            except (
                FileNotFoundError,
                ConnectionError,
                DatasetsError,
                RepositoryNotFoundError,
                requests.ConnectionError,
                requests.ReadTimeout,
            ):
                try:
                    with no_terminal_output():
                        dataset = load_dataset(
                            path=dataset_config.source.split("::")[0],
                            name=(
                                dataset_config.source.split("::")[1]
                                if "::" in dataset_config.source
                                else None
                            ),
                            cache_dir=cache_dir,
                            token=get_hf_token(api_key=api_key),
                        )
                    break
                except (
                    FileNotFoundError,
                    ConnectionError,
                    DatasetsError,
                    RepositoryNotFoundError,
                    requests.ConnectionError,
                    requests.ReadTimeout,
                ) as e:
                    log(
                        f"Failed to load dataset {dataset_config.source!r}, due to "
                        f"the following error: {e}. Retrying...",
                        level=logging.DEBUG,
                    )
                    time.sleep(1)
                    continue
            except HfHubHTTPError:
                raise HuggingFaceHubDown()
        else:
            raise InvalidBenchmark(
                f"Failed to load dataset {dataset_config.source!r} after "
                f"{num_attempts} attempts. Run with verbose mode to see the individual "
                "errors."
            )

    # Case where the dataset source is a dictionary with keys "train", "val" and "test",
    # with the values pointing to local CSV files
    else:
        split_mapping = dict(
            train=dataset_config.train_split,
            val=dataset_config.val_split,
            test=dataset_config.test_split,
        )
        data_files = {
            config_split: dataset_config.source[source_split]
            for source_split, config_split in split_mapping.items()
            if source_split in dataset_config.source and config_split is not None
        }

        # Get the file extension and ensure that all files have the same extension
        file_extensions = {
            config_split: dataset_config.source[source_split].split(".")[-1]
            for source_split, config_split in split_mapping.items()
            if source_split in dataset_config.source and config_split is not None
        }
        if len(set(file_extensions.values())) != 1:
            raise InvalidBenchmark(
                "All data files in a custom dataset must have the same file extension. "
                f"Got the extensions {', '.join(file_extensions.values())} for the "
                f"dataset {dataset_config.name!r}."
            )
        file_extension = list(file_extensions.values())[0]

        # Check that the file extension is supported
        if file_extension not in SUPPORTED_FILE_FORMATS_FOR_LOCAL_DATASETS:
            raise InvalidBenchmark(
                "Unsupported file extension for custom dataset. Supported file "
                "extensions are "
                f"{', '.join(SUPPORTED_FILE_FORMATS_FOR_LOCAL_DATASETS)}, but got "
                f"{file_extension!r}."
            )

        # Load the dataset
        with no_terminal_output():
            dataset = load_dataset(
                path=file_extension, data_files=data_files, cache_dir=cache_dir
            )

    assert isinstance(dataset, DatasetDict)
    return DatasetDict(  # pyrefly: ignore[no-matching-overload]
        {
            split: dataset[split]
            for split in [
                dataset_config.train_split,
                dataset_config.val_split,
                dataset_config.test_split,
            ]
            if split is not None
        }
    )
